"""Utilitiy/Support functions for aiobmsble."""

from fnmatch import translate
import re

from bleak.backends.scanner import AdvertisementData

from aiobmsble import AdvertisementPattern
from aiobmsble.basebms import BaseBMS


def advertisement_matches(
    matcher: AdvertisementPattern,
    adv_data: AdvertisementData,
) -> bool:
    """Determine whether the given advertisement data matches the specified pattern.

    Args:
        matcher (AdvertisementPattern): A dictionary containing the matching criteria.
            Possible keys include:
            - "service_uuid" (str): A specific service 128-bit UUID to match.
            - "service_data_uuid" (str): A specific service data UUID to match.
            - "manufacturer_id" (int): A manufacturer ID to match.
            - "manufacturer_data_start" (bytes): A byte sequence that the data should start with.
            - "local_name" (str): A pattern supporting Unix shell-style wildcards to match

        adv_data (AdvertisementData): An object containing the advertisement data to be checked.

    Returns:
        bool: True if the advertisement data matches the specified pattern, False otherwise.

    """
    if (
        service_uuid := matcher.get("service_uuid")
    ) and service_uuid not in adv_data.service_uuids:
        return False

    if (
        service_data_uuid := matcher.get("service_data_uuid")
    ) and service_data_uuid not in adv_data.service_data:
        return False

    if (manufacturer_id := matcher.get("manufacturer_id")) is not None:
        if manufacturer_id not in adv_data.manufacturer_data:
            return False

        if manufacturer_data_start := matcher.get("manufacturer_data_start"):
            if not adv_data.manufacturer_data[manufacturer_id].startswith(
                bytes(manufacturer_data_start)
            ):
                return False

    return not (
        (local_name := matcher.get("local_name"))
        and not re.compile(translate(local_name)).match(adv_data.local_name or "")
    )


def bms_supported(bms: BaseBMS, adv_data: AdvertisementData) -> bool:
    """Determine if the given BMS is supported based on advertisement data.

    Args:
        bms (BaseBMS): The BMS class to check.
        adv_data (AdvertisementData): The advertisement data to match against.

    Returns:
        bool: True if the BMS is supported, False otherwise.

    """
    for matcher in bms.matcher_dict_list():
        if advertisement_matches(matcher, adv_data):
            return True
    return False



"""Package for battery management systems (BMS) via Bluetooth LE."""

from typing import Literal, TypedDict

type BMSvalue = Literal[
    "battery_charging",
    "battery_level",
    "current",
    "power",
    "temperature",
    "voltage",
    "cycles",
    "cycle_capacity",
    "cycle_charge",
    "delta_voltage",
    "problem",
    "runtime",
    "cell_voltages",
    "design_capacity",
    "temp_values",
    "problem_code",
]


class BMSsample(TypedDict, total=False):
    """Dictionary representing a sample of battery management system (BMS) data."""

    battery_charging: bool
    battery_level: int | float  # [%]
    current: float  # [A]
    power: float  # [W]
    temperature: int | float  # [°C]
    voltage: float  # [V]
    cycle_capacity: int | float  # [Wh]
    cycles: int  # [#]
    delta_voltage: float  # [V]
    problem: bool
    runtime: int  # [s]
    # internal
    cell_voltages: list[float]  # [V]
    cycle_charge: int | float  # [Ah]
    design_capacity: int  # [Ah]
    temp_values: list[int | float]  # [°C]
    problem_code: int


class AdvertisementPattern(TypedDict, total=False):
    """Optional patterns that can match Bleak advertisement data."""

    local_name: str  # name pattern that supports Unix shell-style wildcards
    service_uuid: str  # 128-bit UUID that the device must advertise
    service_data_uuid: str  # service data for the service UUID
    manufacturer_id: int  # required manufacturer ID
    manufacturer_data_start: list[int]  # required starting bytes of manufacturer data



"""Test the Daly BMS implementation."""

import asyncio
from collections.abc import Buffer
from typing import Final
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.exc import BleakError
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble.basebms import BMSsample
from aiobmsble.bms.ogt_bms import BMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient, MockRespChar

base_result: BMSsample = {
    "voltage": 45.681,
    "battery_level": 14,
    "cycles": 99,
    "cycle_charge": 8.0,
    "temperature": 21.8,
    "cycle_capacity": 365.448,
    "power": 56.188,
    "problem": False,
}


# all names result in same encryption key for easier testing
@pytest.fixture(params=["SmartBat-A12345", "SmartBat-B12294"])
def ogt_bms_fixture(request) -> str:
    """Return OGT SmartBMS names."""
    return request.param


class MockOGTBleakClient(MockBleakClient):
    """Emulate an OGT BMS BleakClient."""

    KEY = 0x10  # key used for decoding, constants are encrypted with this key!
    RESP_TYPE_A: dict[int, bytearray] = {
        0x02: bytearray(b" U  \x1d\x1a"),  # battery_level: 14
        0x04: bytearray(b'"  # Q\x1d\x1a'),  # cycle_charge: 8.0
        0x08: bytearray(b"'!R\"\x1d\x1a"),  # voltage: 45.681
        0x0C: bytearray(b"(% R\x1d\x1a"),  # temperature: 21.8
        0x10: bytearray(b"(%VV  \x1d\x1a"),  # current: -1.23
        0x18: bytearray(b"'(  \x1d\x1a"),  # runtime: 7200
        0x2C: bytearray(b"&#  \x1d\x1a"),  # cycles: 99
    }
    RESP_TYPE_B: dict[int, bytearray] = {
        0x08: bytearray(b"(% R\x1d\x1a"),  # temperature: 21.8
        0x09: bytearray(b"'!R\"\x1d\x1a"),  # voltage: 45.681
        0x0A: bytearray(b"'R   Q\x1d\x1a"),  # current: 1.23
        0x0D: bytearray(b" U  \x1d\x1a"),  # battery_level: 14
        0x0F: bytearray(b'"  # Q\x1d\x1a'),  # cycle_charge: 8.0
        0x12: bytearray(b"VVVV\x1d\x1a"),  # runtime: 65536 (inf)
        0x17: bytearray(b"&#  \x1d\x1a"),  # cycles: 99
        0x3F: bytearray(b"UQ S\x1d\x1a"),  # 3.306
        0x3E: bytearray(b"U) S\x1d\x1a"),  # 3.300
        0x3D: bytearray(b"U( S\x1d\x1a"),  # 3.304
        0x3C: bytearray(b"U' S\x1d\x1a"),  # 3.303
    }

    async def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str("fff6"):
            assert self._ble_device.name is not None
            if self._ble_device.name[9] == "A":
                assert (
                    bytearray(data)[0:4] == b";BQQ"
                ), "BMS type A command header incorrect."
            else:
                assert (
                    bytearray(data)[0:4] == b";B!&"
                ), "BMS type B command header incorrect."

            reg: Final[int] = int(
                bytearray((bytearray(data)[x] ^ self.KEY) for x in range(4, 6)).decode(
                    encoding="ascii"
                ),
                16,
            )
            assert self._ble_device.name is not None

            if self._ble_device.name[9] == "A" and reg in self.RESP_TYPE_A:
                return bytearray(b";BT<") + bytearray(data)[4:6] + self.RESP_TYPE_A[reg]

            if self._ble_device.name[9] == "B" and reg in self.RESP_TYPE_B:
                return bytearray(b";BT<") + bytearray(data)[4:6] + self.RESP_TYPE_B[reg]

            return bytearray(b";BT<") + bytearray(b"Ubb\x7f\x10")  # Error

        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool = None,  # noqa: RUF013 # same as upstream
    ) -> None:
        """Issue write command to GATT."""
        # await super().write_gatt_char(char_specifier, data, response)
        assert self._notify_callback is not None
        value: Final[bytearray] = await self._response(char_specifier, data)

        asyncio.get_running_loop().call_soon(
            self._notify_callback, MockRespChar(None, lambda: 0), value
        )


class MockInvalidBleakClient(MockOGTBleakClient):
    """Emulate an invalid BleakClient."""

    async def _response(
        self, char_specifier: BleakGATTCharacteristic | int | str | UUID, data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str("fff6"):
            reg: Final[int] = int(
                bytearray((bytearray(data)[x] ^ self.KEY) for x in range(4, 6)).decode(
                    encoding="ascii"
                ),
                16,
            )
            if reg == 0x8:
                return bytearray(b"")

            return bytearray(b"invalid\xf0value")

        return bytearray()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool = None,  # noqa: RUF013 # same as upstream
    ) -> None:
        """Issue write command to GATT."""
        # await super().write_gatt_char(char_specifier, data, response)
        assert self._notify_callback is not None
        value = await self._response(char_specifier, data)

        # test read timeout on register 8 (valid for A and B type BMS)
        if bytearray(data)[4:6] != bytearray(b" ("):
            self._notify_callback(MockRespChar(None, lambda: 0), value)

    async def disconnect(self) -> bool:
        """Mock disconnect to raise BleakError."""
        raise BleakError


async def test_update(
    patch_bleak_client, ogt_bms_fixture: str, reconnect_fixture: bool
) -> None:
    """Test OGT BMS data update."""

    patch_bleak_client(MockOGTBleakClient)

    bms = BMS(
        generate_ble_device("cc:cc:cc:cc:cc:cc", ogt_bms_fixture, None, -73),
        reconnect_fixture,
    )

    result = await bms.async_update()

    # verify all sensors are reported
    if str(ogt_bms_fixture)[9] == "A":
        assert result == base_result | {
            "current": -1.23,
            "power": -56.188,
            "battery_charging": False,
            "runtime": 7200,
        }
    else:
        assert result == base_result | {
            "current": 1.23,
            "delta_voltage": 0.003,
            "power": 56.188,
            "battery_charging": True,
            "cell_voltages": [3.306, 3.305, 3.304, 3.303],
        }

    # query again to check already connected state
    result: BMSsample = await bms.async_update()
    assert bms._client.is_connected is not reconnect_fixture

    await bms.disconnect()


async def test_update_16s(monkeypatch, patch_bleak_client) -> None:
    """Test OGT BMS data update for 16 cell (max possible)."""

    monkeypatch.setattr(
        MockOGTBleakClient,
        "RESP_TYPE_B",
        MockOGTBleakClient.RESP_TYPE_B
        | {
            key: bytearray(f"U{(chr(0x27),'(',')','Q')[key % 4]} S\x1d\x1a", "ascii")
            for key in range(0x30, 0x3C)
        },
    )
    patch_bleak_client(MockOGTBleakClient)

    bms = BMS(
        generate_ble_device("cc:cc:cc:cc:cc:cc", "SmartBat-B12294", None, -73), False
    )

    # verify all sensors are reported
    assert await bms.async_update() == base_result | {
        "current": 1.23,
        "delta_voltage": 0.003,
        "power": 56.188,
        "battery_charging": True,
        "cell_voltages": [
            3.306,
            3.305,
            3.304,
            3.303,
            3.306,
            3.305,
            3.304,
            3.303,
            3.306,
            3.305,
            3.304,
            3.303,
            3.306,
            3.305,
            3.304,
            3.303,
        ],
    }


@pytest.fixture(
    name="wrong_response",
    params=[
        (bytearray(7), "critical_length"),
        (bytearray(b";AT< )'!R\"\x1d\x1a"), "wrong_SOP"),
        (bytearray(b";BT< )'!R\""), "wrong_EOP"),
        (bytearray(b";BT<#RUN S\x1d\x1a"), "invalid_character"),
        (bytearray(b";BT<Ubb\x7f\x10"), "BMS_error"),
        (bytearray(b"invalid\xf0value"), "invalid_value"),
    ],
    ids=lambda param: param[1],
)
def fix_response(request) -> bytearray:
    """Return faulty response frame."""
    return request.param[0]


async def test_invalid_response(
    monkeypatch, patch_bleak_client, patch_bms_timeout, wrong_response: bytearray
) -> None:
    """Test data up date with BMS returning invalid data."""

    patch_bms_timeout("ogt_bms")

    async def patch_resp(
        _self, char_specifier: BleakGATTCharacteristic | int | str | UUID, _data: Buffer
    ) -> bytearray:
        if isinstance(char_specifier, str) and normalize_uuid_str(
            char_specifier
        ) == normalize_uuid_str("fff6"):
            return wrong_response
        raise NotImplementedError("wrong GATT characteristic")

    monkeypatch.setattr(MockOGTBleakClient, "_response", patch_resp)

    patch_bleak_client(MockOGTBleakClient)

    bms = BMS(generate_ble_device("cc:cc:cc:cc:cc:cc", "SmartBat-B12294", None, -73))

    result: BMSsample = {}
    with pytest.raises(TimeoutError):
        result = await bms.async_update()

    assert not result
    await bms.disconnect()


async def test_invalid_bms_type(patch_bleak_client) -> None:
    """Test BMS with invalid type 'C'."""

    patch_bleak_client(MockOGTBleakClient)

    bms = BMS(generate_ble_device("cc:cc:cc:cc:cc:cc", "SmartBat-C12294", None, -73))

    result: BMSsample = await bms.async_update()
    assert not result
    assert bms._client.is_connected
    await bms.disconnect()



"""Base class defintion for battery management systems (BMS)."""

from abc import ABC, abstractmethod
import asyncio
from collections.abc import Callable
import logging
from statistics import fmean
from typing import Any, Final

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.exc import BleakError
from bleak_retry_connector import establish_connection

from aiobmsble import AdvertisementPattern, BMSsample, BMSvalue

KEY_CELL_VOLTAGE: Final[str] = "cell#"  # [V]


class BaseBMS(ABC):
    """Abstract base class for battery management system."""

    TIMEOUT = 5.0
    _MAX_CELL_VOLT: Final[float] = 5.906  # max cell potential
    _HRS_TO_SECS: Final[int] = 60 * 60  # seconds in an hour

    def __init__(
        self,
        logger_name: str,
        ble_device: BLEDevice,
        reconnect: bool = False,
    ) -> None:
        """Intialize the BMS.

        notification_handler: the callback function used for notifications from 'uuid_rx()'
            characteristic. Not defined as abstract in this base class, as it can be both,
            a normal or async function

        Args:
            logger_name (str): name of the logger for the BMS instance (usually file name)
            ble_device (BLEDevice): the Bleak device to connect to
            reconnect (bool): if true, the connection will be closed after each update

        """
        assert (
            getattr(self, "_notification_handler", None) is not None
        ), "BMS class must define _notification_handler method"
        self._ble_device: Final[BLEDevice] = ble_device
        self._reconnect: Final[bool] = reconnect
        self.name: Final[str] = self._ble_device.name or "undefined"
        self._log: Final[logging.Logger] = logging.getLogger(
            f"{logger_name.replace('.plugins', '')}::{self.name}:"
            f"{self._ble_device.address[-5:].replace(':','')})"
        )

        self._log.debug(
            "initializing %s, BT address: %s", self.device_id(), ble_device.address
        )
        self._client: BleakClient = BleakClient(
            self._ble_device,
            disconnected_callback=self._on_disconnect,
            services=[*self.uuid_services()],
        )
        self._data: bytearray = bytearray()
        self._data_event: Final[asyncio.Event] = asyncio.Event()

    @staticmethod
    @abstractmethod
    def matcher_dict_list() -> list[AdvertisementPattern]:
        """Return a list of Bluetooth advertisement matchers."""

    @staticmethod
    @abstractmethod
    def device_info() -> dict[str, str]:
        """Return a dictionary of device information.

        keys: manufacturer, model
        """

    @classmethod
    def device_id(cls) -> str:
        """Return device information as string."""
        return " ".join(cls.device_info().values())

    @staticmethod
    @abstractmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""

    @staticmethod
    @abstractmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""

    @staticmethod
    @abstractmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""

    @staticmethod
    def _calc_values() -> frozenset[BMSvalue]:
        """Return values that the BMS cannot provide and need to be calculated.

        See _add_missing_values() function for the required input to actually do so.
        """
        return frozenset()

    @staticmethod
    def _add_missing_values(data: BMSsample, values: frozenset[BMSvalue]) -> None:
        """Calculate missing BMS values from existing ones.

        Args:
            data: data dictionary with values received from BMS
            values: list of values to calculate and add to the dictionary

        Returns:
            None

        """
        if not values or not data:
            return

        def can_calc(value: BMSvalue, using: frozenset[BMSvalue]) -> bool:
            """Check value to add does not exist, is requested, and needed data is available."""
            return (value in values) and (value not in data) and using.issubset(data)

        cell_voltages: Final[list[float]] = data.get("cell_voltages", [])
        design_capacity: Final[int | float] = data.get("design_capacity", 0)
        battery_level: Final[int | float] = data.get("battery_level", 0)
        voltage: Final[float] = data.get("voltage", 0)
        cycle_charge: Final[int | float] = data.get("cycle_charge", 0)
        current: Final[float] = data.get("current", 0)

        calculations: dict[BMSvalue, tuple[set[BMSvalue], Callable[[], Any]]] = {
            "voltage": ({"cell_voltages"}, lambda: round(sum(cell_voltages), 3)),
            "delta_voltage": (
                {"cell_voltages"},
                lambda: round(max(cell_voltages) - min(cell_voltages), 3),
            ),
            "cycle_charge": (
                {"design_capacity", "battery_level"},
                lambda: (design_capacity * battery_level) / 100,
            ),
            "cycle_capacity": (
                {"voltage", "cycle_charge"},
                lambda: voltage * cycle_charge,
            ),
            "power": ({"voltage", "current"}, lambda: round(voltage * current, 3)),
            "battery_charging": ({"current"}, lambda: current > 0),
            "runtime": (
                {"current", "cycle_charge"},
                lambda: (
                    int(cycle_charge / abs(current) * BaseBMS._HRS_TO_SECS)
                    if current < 0
                    else None
                ),
            ),
            "temperature": (
                {"temp_values"},
                lambda: round(fmean(data.get("temp_values", [])), 3),
            ),
        }

        for attr, (required, calc_func) in calculations.items():
            if can_calc(attr, frozenset(required)):
                data[attr] = calc_func()

        # do sanity check on values to set problem state
        data["problem"] = any(
            [
                data.get("problem", False),
                data.get("problem_code", False),
                voltage <= 0,
                any(v <= 0 or v > BaseBMS._MAX_CELL_VOLT for v in cell_voltages),
                data.get("delta_voltage", 0) > BaseBMS._MAX_CELL_VOLT,
                cycle_charge <= 0,
                battery_level > 100,
            ]
        )

    def _on_disconnect(self, _client: BleakClient) -> None:
        """Disconnect callback function."""

        self._log.debug("disconnected from BMS")

    async def _init_connection(self) -> None:
        # reset any stale data from BMS
        self._data.clear()
        self._data_event.clear()

        await self._client.start_notify(
            self.uuid_rx(), getattr(self, "_notification_handler")
        )

    async def _connect(self) -> None:
        """Connect to the BMS and setup notification if not connected."""

        if self._client.is_connected:
            self._log.debug("BMS already connected")
            return

        self._log.debug("connecting BMS")
        self._client = await establish_connection(
            client_class=BleakClient,
            device=self._ble_device,
            name=self._ble_device.address,
            disconnected_callback=self._on_disconnect,
            services=[*self.uuid_services()],
        )

        try:
            await self._init_connection()
        except Exception as err:
            self._log.info(
                "failed to initialize BMS connection (%s)", type(err).__name__
            )
            await self.disconnect()
            raise

    async def _await_reply(
        self,
        data: bytes,
        char: BleakGATTCharacteristic | int | str | None = None,
        wait_for_notify: bool = True,
    ) -> None:
        """Send data to the BMS and wait for valid reply notification."""

        self._log.debug("TX BLE data: %s", data.hex(" "))
        self._data_event.clear()  # clear event before requesting new data
        await self._client.write_gatt_char(char or self.uuid_tx(), data, response=False)
        if wait_for_notify:
            await asyncio.wait_for(self._wait_event(), timeout=self.TIMEOUT)

    async def disconnect(self) -> None:
        """Disconnect the BMS, includes stoping notifications."""

        if self._client.is_connected:
            self._log.debug("disconnecting BMS")
            try:
                self._data_event.clear()
                await self._client.disconnect()
            except BleakError:
                self._log.warning("disconnect failed!")

    async def _wait_event(self) -> None:
        """Wait for data event and clear it."""
        await self._data_event.wait()
        self._data_event.clear()

    @abstractmethod
    async def _async_update(self) -> BMSsample:
        """Return a dictionary of BMS values (keys need to come from the SENSOR_TYPES list)."""

    async def async_update(self, raw: bool = False) -> BMSsample:
        """Retrieve updated values from the BMS using method of the subclass.

        Args:
            raw (bool): if true, the raw data from the BMS is returned without
                any calculations or missing values added

        Returns:
            BMSsample: dictionary with BMS values

        """
        await self._connect()

        data: BMSsample = await self._async_update()
        if not raw:
            self._add_missing_values(data, self._calc_values())

        if self._reconnect:
            # disconnect after data update to force reconnect next time (slow!)
            await self.disconnect()

        return data


def crc_modbus(data: bytearray) -> int:
    """Calculate CRC-16-CCITT MODBUS."""
    crc: int = 0xFFFF
    for i in data:
        crc ^= i & 0xFF
        for _ in range(8):
            crc = (crc >> 1) ^ 0xA001 if crc % 2 else (crc >> 1)
    return crc & 0xFFFF


def crc_xmodem(data: bytearray) -> int:
    """Calculate CRC-16-CCITT XMODEM."""
    crc: int = 0x0000
    for byte in data:
        crc ^= byte << 8
        for _ in range(8):
            crc = (crc << 1) ^ 0x1021 if (crc & 0x8000) else (crc << 1)
    return crc & 0xFFFF


def crc8(data: bytearray) -> int:
    """Calculate CRC-8/MAXIM-DOW."""
    crc: int = 0x00  # Initialwert für CRC

    for byte in data:
        crc ^= byte
        for _ in range(8):
            crc = (crc >> 1) ^ 0x8C if crc & 0x1 else crc >> 1

    return crc & 0xFF


def crc_sum(frame: bytearray) -> int:
    """Calculate frame CRC."""
    return sum(frame) & 0xFF



"""Example function for package usage."""

import asyncio
import logging
from types import ModuleType

from bleak import BleakScanner
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from bleak.exc import BleakError

from aiobmsble import BMSsample
from aiobmsble.bms import ogt_bms
from aiobmsble.utils import bms_supported

bms_plugins: list[ModuleType] = [ogt_bms]

logging.basicConfig(
    format="%(levelname)s: %(message)s",
    level=logging.INFO,
)
logger: logging.Logger = logging.getLogger(__name__)

logger.info(
    "loaded BMS types: %s", [key.__name__.rsplit(".", 1)[-1] for key in bms_plugins]
)


async def detect_bms() -> None:
    """Query a Bluetooth device based on the provided arguments."""

    logger.info("starting scan...")
    scan_result: dict[str, tuple[BLEDevice, AdvertisementData]] = (
        await BleakScanner.discover(return_adv=True)
    )
    logger.info("%i BT devices in range.", len(scan_result))

    for ble_dev, advertisement in scan_result.values():
        logger.info(
            "%s\nBT device '%s' (%s)\n\t%s",
            "-" * 72,
            ble_dev.name,
            ble_dev.address,
            repr(advertisement).replace(", ", ",\n\t"),
        )
        for bms_module in bms_plugins:
            if bms_supported(bms_module.BMS, advertisement):
                logger.info(
                    "Found matching BMS type: %s",
                    bms_module.__name__.rsplit(".", maxsplit=1)[-1],
                )
                bms = bms_module.BMS(ble_device=ble_dev, reconnect=True)
                try:
                    logger.info("Updating BMS data...")
                    data: BMSsample = await bms.async_update()
                    logger.info("BMS data: %s", repr(data).replace(", ", ",\n\t"))
                except BleakError as ex:
                    logger.error("Failed to update BMS: %s", ex)

    logger.info("done.")


def main() -> None:
    """Entry point for the script to run the BMS detection."""
    asyncio.run(detect_bms())


if __name__ == "__main__":
    main()
